#  Copyright (c) 2024 Federico Busetti <729029+febus982@users.noreply.github.com>
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

from datetime import date
from typing import Any, Dict, Optional, Sequence, Tuple

from yaml import safe_load
from yaml.scanner import ScannerError

from .base import BaseParser
from .types import TYPE_AST


class MADR3Parser(BaseParser):
    @classmethod
    def _get_metadata_and_ast(cls, file: str) -> Tuple[Dict[str, Any], TYPE_AST]:
        lines = file.splitlines()
        separators = [i for i, x in enumerate(lines) if x == "---"]
        if len(separators) < 2:
            raise LookupError("Metadata section not found in file")

        yaml_file = lines[0 : separators[1] :]
        try:
            metadata: Optional[Dict[str, str]] = safe_load("\n".join(yaml_file))
        except ScannerError:
            raise LookupError("Cannot parse metadata section")

        md_file = lines[separators[1] + 1 : :]
        md_ast = cls.parser.parse("\n".join(md_file))

        return metadata or {}, md_ast

    @classmethod
    def _get_date(cls, metadata: dict, ast: TYPE_AST) -> Optional[date]:
        if isinstance(metadata.get("date"), date):
            return metadata["date"]
        else:
            return None

    @classmethod
    def _get_statuses(cls, metadata: dict, ast: TYPE_AST) -> Sequence[str]:
        if metadata.get("status"):
            return (cls._upperfirst(metadata["status"]),)
        else:
            return tuple()

    @classmethod
    def _get_status(cls, metadata: dict, ast: TYPE_AST) -> Optional[str]:
        try:
            return cls._get_statuses(metadata, ast)[0]
        except IndexError:
            return None

    @classmethod
    def _get_deciders(cls, metadata: dict, ast: TYPE_AST) -> Optional[str]:
        return metadata.get("deciders")

    @classmethod
    def _get_consulted(cls, metadata: dict, ast: TYPE_AST) -> Optional[str]:
        return metadata.get("consulted")

    @classmethod
    def _get_informed(cls, metadata: dict, ast: TYPE_AST) -> Optional[str]:
        return metadata.get("informed")



#  Copyright (c) 2024 Federico Busetti <729029+febus982@users.noreply.github.com>
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#
from .madr2 import MADR2Parser
from .madr3 import MADR3Parser
from .madr4 import MADR4Parser
from .nygard import NygardParser



#  Copyright (c) 2024 Federico Busetti <729029+febus982@users.noreply.github.com>
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

import os
import re
from pathlib import Path
from typing import Optional

from mkdocs_macros.plugin import MacrosPlugin

from .factory import get_parser
from .interfaces import TYPE_ADRStyle
from .renderer import Jinja2Renderer

ADR_REGEX = re.compile("^[0-9]{4}-[\\w-]+\\.md*")


def adr_summary(
    env: MacrosPlugin,
    adr_path: str,
    adr_style: TYPE_ADRStyle,
    template_file: Optional[str] = None,
) -> str:
    absolute_path = Path(env.project_dir).joinpath(adr_path)
    parser = get_parser(adr_style)

    documents = [
        parser.parse(absolute_path.joinpath(f), env.conf["docs_dir"])
        for f in os.listdir(absolute_path)
        if os.path.isfile(absolute_path.joinpath(f)) and ADR_REGEX.match(f)
    ]

    return Jinja2Renderer.summary(
        documents=documents,
        mkdocs_base_path=Path(env.project_dir),
        template_file=template_file,
    )



#  Copyright (c) 2024 Federico Busetti <729029+febus982@users.noreply.github.com>
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from datetime import date, datetime
from typing import Any, Dict, Optional, Sequence, Tuple

from .madr3 import BaseParser
from .types import TYPE_AST


class MADR2Parser(BaseParser):
    @classmethod
    def _get_metadata_and_ast(cls, file: str) -> Tuple[Dict[str, Any], TYPE_AST]:
        if file == "":
            raise LookupError("Cannot parse an empty file")
        md_ast = cls.parser.parse(file)

        return cls._get_metadata_from_ast(md_ast), md_ast

    @classmethod
    def _extract_metadata_as_md_string(cls, ast: TYPE_AST) -> Optional[str]:
        title_index = cls._get_title_index(ast)
        content_index = cls._get_first_h2_index(ast)
        if title_index is None or content_index is None:
            raise LookupError(
                "Malformed document: Could not find"
                " headings surrounding metadata section"
            )

        try:
            metadata_list = next(
                x for x in ast[0][title_index:content_index] if x.get("type") == "list"
            )
        except StopIteration:
            # No metadata in the document
            return None

        return cls.renderer.list(metadata_list, ast[1])

    @classmethod
    def _get_metadata_from_ast(cls, ast: TYPE_AST) -> Dict[str, str]:
        metadata: Dict[str, str] = {}
        rendered_metadata = cls._extract_metadata_as_md_string(ast)

        if rendered_metadata is None:
            return metadata

        for item in rendered_metadata.split("\n"):
            if item.startswith("* Status: "):
                metadata["status"] = item[len("* Status: ") :]
            if item.startswith("* Deciders: "):
                metadata["deciders"] = item[len("* Deciders: ") :]
            if item.startswith("* Date: "):
                metadata["date"] = item[len("* Date: ") :]

        return metadata

    @classmethod
    def _get_date(cls, metadata: dict, ast: TYPE_AST) -> Optional[date]:
        if metadata.get("date"):
            return datetime.strptime(metadata["date"], "%Y-%m-%d").date()
        else:
            return None

    @classmethod
    def _get_statuses(cls, metadata: dict, ast: TYPE_AST) -> Sequence[str]:
        if metadata.get("status"):
            return (cls._upperfirst(metadata["status"]),)
        else:
            return tuple()

    @classmethod
    def _get_status(cls, metadata: dict, ast: TYPE_AST) -> Optional[str]:
        try:
            return cls._get_statuses(metadata, ast)[0]
        except IndexError:
            return None

    @classmethod
    def _get_deciders(cls, metadata: dict, ast: TYPE_AST) -> Optional[str]:
        return metadata.get("deciders")



#  Copyright (c) 2024 Federico Busetti <729029+febus982@users.noreply.github.com>
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
import logging
from abc import ABC, abstractmethod
from datetime import date
from pathlib import Path
from typing import Any, Dict, Optional, Sequence, Tuple

from mistune import Markdown, create_markdown
from mistune.renderers.markdown import MarkdownRenderer
from mkdocs_macros import fix_url

from mkdocs_macros_adr_summary.interfaces import ADRDocument, ADRParser

from .exceptions import InvalidFileError
from .types import TYPE_AST


class BaseParser(ADRParser, ABC):
    renderer: MarkdownRenderer = MarkdownRenderer()
    parser: Markdown = create_markdown(
        escape=False,
        renderer=None,
        plugins=["strikethrough", "footnotes", "table", "speedup"],
    )

    @classmethod
    def parse(cls, file_path: Path, base_path: Path) -> ADRDocument:
        with open(file_path, "r") as f:
            file = f.read()
        try:
            metadata, ast = cls._get_metadata_and_ast(file)
        except LookupError as e:
            raise InvalidFileError(file_path) from e

        doc = ADRDocument(
            file_path=fix_url(str(file_path.relative_to(base_path))),
            document_id=cls._get_id(file_path),
            title=cls._get_title(metadata, ast),
            date=cls._get_date(metadata, ast),
            status=cls._get_status(metadata, ast),
            statuses=cls._get_statuses(metadata, ast),
            deciders=cls._get_deciders(metadata, ast),
            consulted=cls._get_consulted(metadata, ast),
            informed=cls._get_informed(metadata, ast),
        )
        return doc

    @classmethod
    def _get_title(cls, metadata: dict, ast: TYPE_AST) -> Optional[str]:
        title_index = cls._get_title_index(ast)
        if title_index is None:
            return None
        return cls.renderer.paragraph(ast[0][title_index], ast[1]).strip()

    @classmethod
    def _get_title_index(cls, ast: TYPE_AST) -> Optional[int]:
        h1_file_lines = [
            i
            for i, x in enumerate(ast[0])
            if x.get("type") == "heading" and x.get("attrs", {}).get("level") == 1
        ]
        if len(h1_file_lines) != 1:
            logging.warning("Malformed content: Could not find the title heading.")
            return None

        return h1_file_lines[0]

    @classmethod
    def _get_first_h2_index(cls, ast: TYPE_AST) -> Optional[int]:
        h2_file_lines = [
            i
            for i, x in enumerate(ast[0])
            if x.get("type") == "heading" and x.get("attrs", {}).get("level") == 2
        ]
        if len(h2_file_lines) < 1:
            logging.warning("Malformed content: Could not find the title heading.")
            return None

        return h2_file_lines[0]

    @classmethod
    @abstractmethod
    def _get_date(cls, metadata: dict, ast: TYPE_AST) -> Optional[date]: ...

    @classmethod
    @abstractmethod
    def _get_statuses(cls, metadata: dict, ast: TYPE_AST) -> Sequence[str]: ...

    @classmethod
    @abstractmethod
    def _get_status(cls, metadata: dict, ast: TYPE_AST) -> Optional[str]: ...

    @classmethod
    @abstractmethod
    def _get_metadata_and_ast(cls, file: str) -> Tuple[Dict[str, Any], TYPE_AST]: ...

    @classmethod
    def _get_id(cls, file_path: Path) -> Optional[int]:
        try:
            return int(file_path.parts[-1][0:4])
        except ValueError:
            return None

    @classmethod
    def _get_deciders(cls, metadata: dict, ast: TYPE_AST) -> Optional[str]:
        return None

    @classmethod
    def _get_consulted(cls, metadata: dict, ast: TYPE_AST) -> Optional[str]:
        return None

    @classmethod
    def _get_informed(cls, metadata: dict, ast: TYPE_AST) -> Optional[str]:
        return None

    @staticmethod
    def _upperfirst(text: str) -> str:
        return text[0].capitalize() + text[1::] if text[0].islower() else text



#  Copyright (c) 2024 Federico Busetti <729029+febus982@users.noreply.github.com>
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

from typing import Dict, Type

from .interfaces import ADRParser, TYPE_ADRStyle
from .parser import MADR2Parser, MADR3Parser, MADR4Parser, NygardParser

parser_registry: Dict[TYPE_ADRStyle, Type[ADRParser]] = {
    "nygard": NygardParser,
    "MADR4": MADR4Parser,
    "MADR3": MADR3Parser,
    "MADR2": MADR2Parser,
}


def get_parser(adr_style: TYPE_ADRStyle) -> Type[ADRParser]:
    try:
        parser = parser_registry[adr_style]
    except KeyError:
        raise ValueError(f"Format {adr_style} not supported")

    return parser



#  Copyright (c) 2024 Federico Busetti <729029+febus982@users.noreply.github.com>
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

from datetime import date, datetime
from typing import Any, Dict, List, Optional, Sequence, Tuple

from .base import BaseParser
from .types import TYPE_AST


class NygardParser(BaseParser):
    @classmethod
    def _get_metadata_and_ast(cls, file: str) -> Tuple[Dict[str, Any], TYPE_AST]:
        md_ast = cls.parser.parse(file)
        return {}, md_ast

    @classmethod
    def _get_date(cls, metadata: dict, ast: TYPE_AST) -> Optional[date]:
        title_index = cls._get_title_index(ast)
        if title_index is None:
            return None

        try:
            block = ast[0][title_index + 2]
        except IndexError:
            return None

        if not block.get("type") == "paragraph":
            return None

        raw_text = cls.renderer.paragraph(block, ast[1]).strip()
        try:
            return datetime.strptime(raw_text, "Date: %Y-%m-%d").date()
        except ValueError:
            return None

    @classmethod
    def _get_statuses(cls, metadata: dict, ast: TYPE_AST) -> Sequence[str]:
        statuses: List[str] = []

        # Find status header
        h2_list = [
            i
            for i, x in enumerate(ast[0])
            if x.get("type") == "heading"
            and x.get("attrs", {}).get("level") == 2
            and cls.renderer.paragraph(x, ast[1]).strip() == "Status"
        ]
        if len(h2_list) != 1:
            return tuple(statuses)

        i = h2_list[0] + 1
        while i < len(ast[0]):
            block = ast[0][i]

            if block.get("type") == "paragraph":
                statuses.append(cls.renderer.paragraph(block, ast[1]).strip())
                i += 1
            elif block.get("type") == "heading":
                break
            else:
                i += 1
                continue

        return tuple(statuses)

    @classmethod
    def _get_status(cls, metadata: dict, ast: TYPE_AST) -> Optional[str]:
        statuses = cls._get_statuses(metadata, ast)
        return statuses[-1] if statuses else None
